/*
 *  $Id: dmnx.h,v 1.4 1993/11/12 11:54:52 doss Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

/* 
   These are the NX-specific macros.

   This file also contains some information on the design of the NX/MPI
   transport layer.
 */

/*D
      NX-MPI transport (device) implementation

      This directory contains the NX implementation of the MPI device layer.
      It is made from the NX device layer.
      This is a simple implementation, providing only the most rudimentary
      functions.  This is an example of how simple an MPI implemenation can be,
      with MPI code from the reference implementation handling the details of
      communicators, derived data types, and other things.  
      
      We only assume that we can send and receive packets of data between
      processors.  We DO assume that a non-trivial amount of buffering is
      provided by the message-passing system.
D*/

#ifndef _DMNX_INCLUDED
#define _DMNX_INCLUDED

#include "dmpiatom.h"

#define MPID_PACKET_SIZE 100000


/*
  The implementation reserves some message tags.

  (An optimization is to allow the use of all but a few very large tags
  for messages in the initial communicator, thus eliminating a separate
  header.  Messages in a different communicator would be sent on a reserved
  set of tags.  An alternate is to use the NX tags for communicator
  types, making the message-passing system handle the queueing of messages
  by communicator

  PT2PT_TAG is the tag for short messages and the headers of long messages
  PT2PT2_TAG(src) is the tag for longer messages (by source).  This permits
  the header messages to be freely received into preallocated buffers, and
  for long messages to be received directly into user-buffers.
 */
#define MPID_PT2PT_TAG 0
#define MPID_PT2PT2_TAG(src) (1+(src))

/* Whether an operation should block or not */
typedef enum { MPID_NOTBLOCKING = 0, MPID_BLOCKING } MPID_BLOCKING_TYPE;
typedef struct {
    int len,			/* TOTAL length of message in BYTES */
        tag,			/* Message tag */
        context_id;		/* Internal communicator ID */
    char buffer[MPID_PACKET_SIZE];
				/* Minimum message size */
    } MPID_PACKET;
/* HeaderLen is just the length of the envelope of MPID_PACKET */
#define MPID_HEADER_LEN (sizeof(MPID_PACKET)-MPID_PACKET_SIZE)

/*
   Another option would be for the device handle to contain the initial
   packet.  I have NOT do this so as to keep down the size of the device
   handle (since I want relatively large packets)

   This needs to be structured so that there is just enough here for mpir.h.
 */
typedef struct {
    int           to, context_id, tag;
    int           rel_rank;              /* sender's rank in group */
    MPIR_BOOL     done;
    char          *buffer;
    int		   sid;
        /* The following describes the buffer to be sent */
    void          *start;
    int           count;
    int           bytes_as_contig;
    MPIR_NODETYPE dataelement;
    } MPID_SHANDLE;

typedef struct {
    int           from, context_id, tag, len;
    MPIR_BOOL     done;
    int           is_posted;        /* set if in the posted receive queue */
    char          *temp;
    int		  rid;
        /* The following describes the buffer to be recv */
    void          *start;
    int           count;
    int           bytes_as_contig;
    MPIR_NODETYPE dataelement;
    } MPID_RHANDLE;

#define MPID_MIN(a,b) ((a) < (b) ? (a) : (b))

#define MPID_CAN_SEND_CONTIG

/* Error values come from dmpiatom.h */

/* 
   Since allocation is done by placing the device structure directly into
   the MPIR_?HANDLE, we don't need to allocate space.  We do, however, take
   this opportunity to initialize it...
 */
#define MPID_alloc_send_handle( a ) {(a)->done = NO;(a)->buffer= 0;}
#define MPID_alloc_recv_handle( a ) {(a)->done = NO;(a)->temp  = 0;}
#define MPID_free_send_handle( a )  if ((a)->buffer) {FREE((a)->buffer);}
#define MPID_free_recv_handle( a )  if ((a)->temp  ) {FREE((a)->temp);}
#ifdef FOO
#define MPID_address_send_handle( a ) &(a)
#define MPID_address_recv_handle( a ) &(a)

/*
   Access routines for the device's part of the structure
 */
#define MPID_set_send_tag( a, v )  (a)->tag    = v
#define MPID_set_send_context( a, v ) (a)->context_id   = (int)(v)
#define MPID_set_send_dest( a, v ) (a)->to     = v
#define MPID_set_send_start( a, v ) (a)->start = v
#define MPID_set_send_count( a, v ) (a)->count = v
#define MPID_set_send_dataelement( a, v ) (a)->dataelement = v
#define MPID_set_send_source_in_group( a, v ) (a)->rel_rank = v
#define MPID_set_recv_tag( a, v )  (a)->tag    = v
#define MPID_set_recv_context( a, v ) (a)->context_id   = (int)(v)
#define MPID_set_recv_source( a, v ) (a)->from = v
#define MPID_set_recv_start( a, v ) (a)->start = v
#define MPID_set_recv_count( a, v ) (a)->count = v
#define MPID_set_recv_dataelement( a, v ) (a)->dataelement = v
#endif

/* Contact with the device layer is made here.  These call the
   routines to actually process a message (see dmnx.c) */
#define MPID_post_send(dmpi_send_handle) \
    MPID_NX_post_send(dmpi_send_handle) 
#define MPID_post_recv(dmpi_recv_handle, is_available ) \
    MPID_NX_post_recv(dmpi_recv_handle, is_available ) 
#define MPID_complete_recv(dmpi_recv_handle) \
    MPID_NX_complete_recv(dmpi_recv_handle) 
#define MPID_complete_send(dmpi_send_handle,status) \
    MPID_NX_complete_send(dmpi_send_handle,status) 
#define MPID_check_device( blocking ) \
    MPID_NX_check_device( blocking )

#ifdef FOO
#define MPID_check_for_unexpected( source_dest,tag,context_id,found,handleptr)\
    MPIR_search_queue( &unexpected_recvs, source_dest, tag, context_id, found,\
		       1,handleptr )
#endif

extern void MPID_NX_myrank(), MPID_NX_mysize(), MPID_NX_End();
#define MPID_myrank( rank ) MPID_NX_myrank( rank )
#define MPID_mysize( size ) MPID_NX_mysize( size )
#define MPID_end()          MPID_NX_End()

#define _NX_me		    _mynode()

#endif

#ifdef MPID_DEVICE_CODE
#include "../../include/dmpiatom.h"
#endif

